--------------------------------------------------------------------------------
--! @File        : CipherCore.vhd
--! @Brief       : Cipher core
--!   ______   ________  _______    ______
--!  /      \ /        |/       \  /      \
--! /$$$$$$  |$$$$$$$$/ $$$$$$$  |/$$$$$$  |
--! $$ |  $$/ $$ |__    $$ |__$$ |$$ | _$$/
--! $$ |      $$    |   $$    $$< $$ |/    |
--! $$ |   __ $$$$$/    $$$$$$$  |$$ |$$$$ |
--! $$ \__/  |$$ |_____ $$ |  $$ |$$ \__$$ |
--! $$    $$/ $$       |$$ |  $$ |$$    $$/
--!  $$$$$$/  $$$$$$$$/ $$/   $$/  $$$$$$/
--!
--! @Author      : Farnoud Farahmand and Panasayya Yalla
--! @Copyright   : Copyright © 2016 Cryptographic Engineering Research Group
--!                ECE Department, George Mason University Fairfax, VA, U.S.A.
--!                All rights Reserved.
--------------------------------------------------------------------------------
--! Description
--!
--!
--!
--!
--!
--!
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use ieee.std_logic_unsigned.all;
use IEEE.math_real."ceil";
use IEEE.math_real."log2";
use work.GeneralComponents_pkg.ALL;
use work.Design_pkg.all;
use work.CAESAR_LWAPI_pkg.all;

entity CipherCore is
    generic (
        G_DBLK_SIZE     : integer := 128
    );
    Port (
            clk             : in   STD_LOGIC;
            rst             : in   STD_LOGIC;
            --PreProcessor===============================================
            ----!key----------------------------------------------------
            key             : in   STD_LOGIC_VECTOR (SW      -1 downto 0);
            key_valid       : in   STD_LOGIC;
            key_ready       : out  STD_LOGIC;
            ----!Data----------------------------------------------------
            bdi             : in   STD_LOGIC_VECTOR (PW      -1 downto 0);
            bdi_valid       : in   STD_LOGIC;
            bdi_ready       : out  STD_LOGIC;
            bdi_partial     : in   STD_LOGIC;
            bdi_pad_loc     : in   STD_LOGIC_VECTOR (PWdiv8  -1 downto 0);
            bdi_valid_bytes : in   STD_LOGIC_VECTOR (PWdiv8  -1 downto 0);
            bdi_size        : in   STD_LOGIC_VECTOR (3       -1 downto 0);
            bdi_eot         : in   STD_LOGIC;
            bdi_eoi         : in   STD_LOGIC;
            bdi_type        : in   STD_LOGIC_VECTOR (4       -1 downto 0);
            decrypt_in      : in   STD_LOGIC;
            key_update      : in   STD_LOGIC;
            --!Post Processor=========================================
            bdo             : out  STD_LOGIC_VECTOR (PW      -1 downto 0);
            bdo_valid       : out  STD_LOGIC;
            bdo_ready       : in   STD_LOGIC;
            bdo_type        : out  STD_LOGIC_VECTOR (4       -1 downto 0);
            bdo_valid_bytes : out  STD_LOGIC_VECTOR (PWdiv8  -1 downto 0);
            end_of_block    : out  STD_LOGIC;
            decrypt_out     : out  STD_LOGIC;
            msg_auth_valid  : out  STD_LOGIC ;
            msg_auth_ready  : in   STD_LOGIC ;
            msg_auth        : out  STD_LOGIC
         );

end CipherCore;

architecture structure of CipherCore is
   constant  NumOfWords : integer:=G_DBLK_SIZE/PW;
   constant  NumOfNWords: integer:= 96/PW;
   constant  addrWidth  : integer:=integer(ceil(log2(real(NumOfWords))));
   signal    dout_kram  : std_logic_vector(PW         -1 downto 0);
   signal    addr_kram  : std_logic_vector(addrWidth  -1 downto 0);
   signal    wen_kram   : std_logic;
   signal    dout_npub  : std_logic_vector(PW         -1 downto 0);
   signal    addr_npub  : std_logic_vector(addrWidth  -1 downto 0);
   signal    wen_npub   : std_logic;
   signal    dout_ad    : std_logic_vector(PW         -1 downto 0);
   signal    addr_ad    : std_logic_vector(addrWidth  -1 downto 0);
   signal    wen_ad     : std_logic;
   signal    dout_data  : std_logic_vector(PW         -1 downto 0);
   signal    addr_data  : std_logic_vector(addrWidth  -1 downto 0);
   signal    wen_data   : std_logic;
   signal    dout_cdata : std_logic_vector(PW         -1 downto 0);
   signal    din_cdata  : std_logic_vector(PW         -1 downto 0);
   signal    addr_cdata : std_logic_vector(addrWidth  -1 downto 0);
   signal    wen_cdata  : std_logic;
   signal    dout_rnd   : std_logic_vector(PW         -1 downto 0);
   signal    din_Tdata  : std_logic_vector(PW         -1 downto 0);
   signal    addr_Tdata : std_logic_vector(addrWidth  -1 downto 0);
   signal    wen_Tdata  : std_logic;
   signal    dout_tag   : std_logic_vector(PW         -1 downto 0);
   signal    tag        : std_logic_vector(PW         -1 downto 0);
   signal    zcount     : integer;
   signal    wrd_count : std_logic_vector(addrWidth -1 downto 0);
   signal    partial_blk_size : std_logic_vector(addrWidth -1 downto 0);
   signal    rnd_count : std_logic_vector(32 -1 downto 0);
   signal    rst_wcnt  : std_logic;
   signal    en_wcnt   : std_logic;
   signal    rst_rcnt  : std_logic;
   signal    en_rcnt   : std_logic;
   signal    zdone     : std_logic;
   signal    zpdone    : std_logic;
   signal    ndone     : std_logic;
   signal    sel_tag   : std_logic;
   signal    bdo_valid_internal: std_logic;
   constant  rnd_const : std_logic_vector(PW         -1 downto 0):=(others=>'1');
   signal    bdi_ready_s    : std_logic;
   signal    bdi_valid_s    : std_logic;
   signal    bdi_eot_pad    : std_logic;
   signal    end_of_block_s : std_logic;
   signal    bdi_padded     : std_logic_vector(PW    -1 downto 0);
   signal    rnd_count_BS   : std_logic_vector(G_DBLK_SIZE      -1 downto 0);
   signal    rnd_i          : std_logic_vector(PW    -1 downto 0);
   type word_array is array (0 to NumOfWords-1) of std_logic_vector(PW-1 downto 0);
   signal rnd_i_array       : word_array;
   signal len_i_array       : word_array;
   signal len_ad            : std_logic_vector(32               -1 downto 0);
   signal len_ad_bits       : std_logic_vector(32               -1 downto 0);
   signal len_d             : std_logic_vector(32               -1 downto 0);
   signal len_d_bits        : std_logic_vector(32               -1 downto 0);
   constant ZEROS           : std_logic_vector(G_DBLK_SIZE      -1 downto 0)
                                := (others => '0');
   signal len               : std_logic_vector(G_DBLK_SIZE      -1 downto 0);
   signal sel_npub_din      : std_logic;
   signal end_of_block_in_s : std_logic;
   signal sel_npub          : std_logic;
   signal nram_din          : std_logic_vector(PW    -1 downto 0);
   signal m_c               : std_logic_vector(PW    -1 downto 0);
   signal bdi_eot_next      : std_logic;
   signal bdi_eot_r         : std_logic;
   signal bdi_eoi_next      : std_logic;
   signal bdi_eoi_r         : std_logic;
   signal tag_error_next    : std_logic;
   signal tag_error_r       : std_logic;
   signal rst_pad           : std_logic;
   signal clr_accum         : std_logic;
   signal partial_next      : std_logic;
   signal partial_r         : std_logic;
   signal zdone_ct          : std_logic;
   signal pad_block         : std_logic;
   signal pad_blk_proc_next : std_logic;
   signal pad_blk_proc_r    : std_logic;
   signal bdo_s             : std_logic_vector(PW    -1 downto 0);
   signal bdo_padded        : std_logic_vector(PW    -1 downto 0);
   signal round_padded      : std_logic_vector(PW    -1 downto 0);
   signal round             : std_logic_vector(PW    -1 downto 0);
   signal en_selMode        : std_logic;
   signal len_selMode       : std_logic;
   signal dout_selMode      : std_logic;
   signal temp              : std_logic_vector(PW    -1 downto 0);
   --! debugging
   signal test_rnd_pad : std_logic;

   --!control unit
    type t_state is (S_INIT, S_LD_KEY, S_LD_NPUB, S_LD_AD, S_LD_PT, S_OUT_CT, S_OUT_TAG,
                    S_PROC_DATA,S_PROC_TAG, S_VERIFY_TAG,
                    S_WAIT_KEY, S_WAIT_NPUB, S_WAIT_AD, S_WAIT_PT,
                    S_DONE);
    signal nx_state, pr_state:t_state;
begin

    --  =======================================================================
    --  Padding unit
    --  =======================================================================
        padding_unit: entity work.padding_unit(behavioral)
            port map(
                    rst             => rst_pad,
                    clk             => clk,
                    bdi_in          => bdi,
                    bdi_valid_in    => bdi_valid,
                    bdi_ready_in    => bdi_ready_s,
                    bdo_in          => bdo_s,
                    round_in        => round,
                    bdi_out         => bdi_padded,
                    bdi_valid_out   => bdi_valid_s,
                    bdi_ready_out   => bdi_ready,
                    bdo_out         => bdo_padded,
                    round_out       => round_padded,
                    bdi_eot_in      => bdi_eot,
                    bdi_eot_out     => bdi_eot_pad,
                    bdi_pad_loc     => bdi_pad_loc,
                    bdi_valid_bytes => bdi_valid_bytes,
                    end_of_block    => end_of_block_in_s,
                    bdi_type        => bdi_type,
                    pad_block       => pad_block
            );
        -- DEBUG
        --bdi_padded    <= bdi; --Test
        -------------
    --! =======================================================================
    --! Datapath
    --! =======================================================================
    --L1: for i in NumOfWords-1 downto 0 generate
        KRAM: SPDRam
        generic map(PW,addrWidth)
        port map
            (
                clk  =>clk,
                wen  =>wen_kram,
                addr =>addr_kram,
                din  =>key,
                dout =>dout_kram
            );
        NRAM: SPDRam
        generic map(PW,addrWidth)
        port map
            (
                clk  =>clk,
                wen  =>wen_npub,
                addr =>addr_npub,
                din  =>nram_din,
                dout =>dout_npub
            );

        ARAM: SPDRam
        generic map(PW,addrWidth)
        port map
            (
                clk  =>clk,
                wen  =>wen_ad,
                addr =>addr_ad,
                din  =>bdi_padded,
                dout =>dout_ad
            );
        DRAM: SPDRam
        generic map(PW,addrWidth)
        port map
            (
                clk  =>clk,
                wen  =>wen_data,
                addr =>addr_data,
                din  =>bdi_padded,
                dout =>dout_data
            );
        CRAM: SPDRam
        generic map(PW,addrWidth)
        port map
            (
                clk  =>clk,
                wen  =>wen_cdata,
                addr =>addr_cdata,
                din  =>din_cdata,
                dout =>dout_cdata
            );

        TRAM: SPDRam
        generic map(PW,addrWidth)
        port map
            (
                clk  =>clk,
                wen  =>wen_Tdata,
                addr =>addr_Tdata,
                din  =>din_Tdata,
                dout =>dout_tag
            );

    --end generate;


    rnd_count_BS  <= ZEROS(G_DBLK_SIZE -1 downto 32) & rnd_count(31 downto 0);

    round_counter: for i in 0 to NumOfWords-1 generate
        rnd_i_array(i) <= rnd_count_BS(G_DBLK_SIZE - (PW*i)-1 downto G_DBLK_SIZE - (PW*i)-PW);
    end generate;

    len_ad_bits <= len_ad (28 downto 0) & "000";
    len_d_bits  <= len_d (28 downto 0) & "000";
    len <= ZEROS(G_DBLK_SIZE-1 downto 96) & len_ad_bits & ZEROS(63 downto 32) & len_d_bits;

    Len_gen: for i in 0 to NumOfWords-1 generate
        len_i_array(i) <= len(G_DBLK_SIZE - (PW*i)-1 downto G_DBLK_SIZE - (PW*i)-PW);
    end generate;

    m_c           <= din_cdata when ((dout_SelMode='1') and (pr_state /= S_LD_AD)) else bdi_padded; --and (pad_block='0')
    round         <= dout_kram xor dout_npub xor rnd_i_array(to_integer(unsigned(wrd_count))) xor dout_data;
    dout_rnd      <= round_padded when (pad_blk_proc_r='0')and(zpdone='1') else
                     round        when pad_blk_proc_r='0' else
                     dout_data;
    temp <=rnd_i_array(to_integer(unsigned(wrd_count)));
    --! debugging
    test_rnd_pad <= '1' when(pad_blk_proc_r='0')and(zpdone='1') else '0';

    din_Tdata     <= nram_din   when sel_npub = '1' else (m_c xor dout_tag);
    din_cdata     <= tag        when sel_tag  = '1' else dout_rnd;
    tag           <= dout_kram xor len_i_array(to_integer(unsigned(wrd_count))) xor dout_tag;
    bdo_s         <= dout_cdata when bdo_valid_internal='1' else (others=>'0');
    bdo           <= bdo_padded when (pr_state = S_OUT_CT)and(end_of_block_s='1') else bdo_s;
    end_of_block  <= end_of_block_s;

    end_of_block_in_s <= ndone when (sel_npub = '1') else zdone;
    nram_din          <= ZEROS(PW-1 downto 0) when (sel_npub_din = '1') else bdi_padded;
    --! =======================================================================
    --! Control
    --! =======================================================================
    --=====STATE REGISTER=======================================================
    process (clk,rst)
    begin
        if rising_edge(clk) then
			if(rst='1')  then
				pr_state <= S_INIT;
				len_ad <= (others => '0');
				len_d  <= (others => '0');
				partial_blk_size <= (others => '0');
				partial_next <= '0';
				--pad_blk_proc_next <= '0';
			else
				pr_state <= nx_state;
				if clr_accum = '1' then
				    len_ad <= (others => '0');
				    len_d  <= (others => '0');
				    partial_blk_size <= (others => '0');
				    partial_next <= '0';
				elsif ((bdi_ready_s = '1') and (bdi_valid = '1') and (pad_block = '0')) then
				    if (bdi_type = HDR_AD) then
				        len_ad <= std_logic_vector(unsigned(len_ad) + unsigned(bdi_size));
				    elsif (bdi_type = HDR_MSG) then
				        len_d <= std_logic_vector(unsigned(len_d) + unsigned(bdi_size));
				        if (bdi_eot = '1') and (end_of_block_s = '0') then
				            partial_next <= '1';
				            partial_blk_size <= wrd_count;
				        end if;
				    end if;
				end if;
				bdi_eot_r <= bdi_eot_next;
				bdi_eoi_r <= bdi_eoi_next;
				tag_error_r <= tag_error_next;
				partial_r <= partial_next;
				pad_blk_proc_r <= pad_blk_proc_next;
			end if;
        end if;
    end process;
    --==========================================================================
    --!next state function
    --==========================================================================
    process(pr_state,key_valid, bdi_valid_s, bdo_ready,bdi_pad_loc, bdi_valid_bytes,bdi_size, bdi_valid,
            bdi_eot, bdi_eot_r, bdi_eoi, bdi_eoi_r, bdi_type, dout_SelMode, key_update, zdone,zdone_ct, pad_block)
    begin
        nx_state <= pr_state;
        case pr_state is
            when S_INIT=>
                nx_state <= S_WAIT_KEY;
                --if (bdi_valid='1')then
                    -- if (key_update='1')then
                        -- nx_state <= S_WAIT_KEY;
                    -- else
                        -- nx_state <= S_WAIT_NPUB;
                    -- end if;
                -- end if;

            when S_WAIT_KEY=>
                if(key_valid='1' )then --and bdi_type=HDR_KEY
                    nx_state <= S_LD_KEY;
                elsif(bdi_valid='1') then
                    nx_state <= S_WAIT_NPUB;
                else
                    nx_state <= S_WAIT_KEY;
                end if;

            when S_LD_KEY=>
                if(key_valid='1' and zdone='1') then
                    nx_state <= S_WAIT_NPUB;
                else
                    nx_state <= S_LD_KEY;
                end if;

            when S_WAIT_NPUB=>
                if(bdi_valid_s='1' and bdi_type=HDR_NPUB)then
                    nx_state <= S_LD_NPUB;
                else
                    nx_state <= S_WAIT_NPUB;
                end if;


            when S_LD_NPUB=>
                if(zdone='1') then -- bdi_valid_s='1' and
                    if ((bdi_eoi_r='1') or (bdi_eoi='1')) then
                        nx_state <= S_PROC_TAG;
                    else
                        nx_state <= S_WAIT_AD;
                    end if;
                else
                    nx_state <= S_LD_NPUB;
                end if;

            when S_WAIT_AD=>
                if(bdi_valid_s='1')then --and bdi_type=HDR_AD
                    if bdi_type=HDR_AD then
                        nx_state <= S_LD_AD;
                    else
                        nx_state <= S_WAIT_PT;
                    end if;
                else
                    nx_state <= S_WAIT_AD;
                end if;

            when S_LD_AD=>
                if(bdi_valid_s='1' and zdone='1') then
                    if ((bdi_eot_r='1') or (bdi_eot='1')) then
                        if (bdi_eoi_r='1') or ((bdi_eoi='1') and (pad_block = '0')) then
                            nx_state <= S_PROC_TAG;
                        else
                            nx_state <= S_WAIT_PT;
                        end if;
                    else
                        nx_state <= S_WAIT_AD;
                    end if;
                else
                    nx_state <= S_LD_AD;
                end if;

            when S_WAIT_PT=>
                if(bdi_valid_s='1' and bdi_type=HDR_MSG)then
                    nx_state <= S_LD_PT;
                else
                    nx_state <= S_WAIT_PT;
                end if;

            when S_LD_PT=>
                if(bdi_valid_s='1' and zdone='1') then
                    nx_state <= S_PROC_DATA;
                else
                    nx_state <= S_LD_PT;
                end if;

            when S_PROC_DATA=>
                if(zdone='1') then
                    nx_state <= S_OUT_CT;
                else
                    nx_state <= S_PROC_DATA;
                end if;

            when S_OUT_CT=>
                if (zdone_ct='1') and (bdo_ready='1')then
                    if bdi_eot_r='1' then --or (bdi_eot='1')
                        nx_state <= S_PROC_TAG;
                    else
                        nx_state <= S_WAIT_PT;
                    end if;
                else
                    nx_state <= S_OUT_CT;
                end if;

            when S_PROC_TAG=>
                if(zdone='1') then
                    --if(dout_SelMode='1') then
                      --  nx_state <= S_VERIFY_TAG;
                    --else
                        nx_state <= S_OUT_TAG;
                    --end if;
                else
                    nx_state <= S_PROC_TAG;
                end if;

            when S_OUT_TAG=>
                if(zdone='1' and bdo_ready='1') then
                    nx_state <= S_DONE;
                else
                    nx_state <= S_OUT_TAG;
                end if;

            when S_VERIFY_TAG=>
                if(zdone='1') then
                    nx_state <= S_DONE;
                end if;

            when S_DONE=>
                nx_state <= S_INIT;

            when others=>
                nx_state <= S_INIT;

        end case;
    end process;
    --==========================================================================
    --!output function
    --==========================================================================
    process(pr_state,key_valid, bdi_valid_s, bdi_valid, bdo_ready, bdi_pad_loc, bdi_valid_bytes,
            bdi_size, bdi_eot,bdi_eoi, bdi_type, dout_SelMode, key_update, zcount, pad_block, bdi, dout_tag,
            zdone, zpdone, zdone_ct,tag_error_r, pad_blk_proc_r, partial_r, bdi_eot_r, bdi_eoi_r)


    begin
            key_ready   <='0';
            bdi_ready_s   <='0';
            bdo_valid_internal   <='0';
            sel_tag     <='0';
            wen_kram    <='0';
            wen_npub    <='0';
            wen_ad      <='0';
            wen_data    <='0';
            wen_cdata   <='0';
            wen_Tdata   <='0';
            rst_wcnt    <='1';
            rst_rcnt    <='1';
            en_SelMode  <='0';
            len_SelMode <='0';
            --rst_tcnt    <='1';
            en_wcnt     <='0';
            en_rcnt     <='0';
            --en_tcnt     <='0';
            end_of_block_s<='0';
            bdo_type    <= (others=>'0');
            --bdo_size    <= (others=>'0');
            msg_auth_valid<='0';
            msg_auth<='0';
            bdo_valid_bytes<=(others=>'1');
            sel_npub_din  <= '0';
            sel_npub      <= '0';
            --done <='0';
            bdi_eot_next  <= bdi_eot_r; --to avoid latch
            bdi_eoi_next  <= bdi_eoi_r; --to avoid latch
            tag_error_next <= tag_error_r; --to avoid latch
            pad_blk_proc_next <= pad_blk_proc_r; --to avoid latch
            rst_pad<='0';
            clr_accum <= '0';

        case pr_state is

            when S_INIT=>
                rst_wcnt                  <= '0';
                rst_rcnt                  <= '0';
                --rst_tcnt                <= '0';
                bdi_eot_next              <= '0';
                bdi_eoi_next              <= '0';
                tag_error_next            <= '0';
                rst_pad                   <='1';
                clr_accum                 <= '1';
                pad_blk_proc_next         <= '0';
                len_SelMode               <='1';

            when S_WAIT_KEY=>
                rst_pad                   <='1';

            when S_LD_KEY=>
                key_ready                 <='1';
                wen_kram                  <=key_valid;
                en_wcnt                   <=key_valid;
                rst_wcnt                  <=not(key_valid and zdone);
                rst_pad                   <='1';

            when S_WAIT_NPUB=>
                -- bdi_ready_s            <='1';
                rst_pad                   <='1';

            when S_LD_NPUB=>
                sel_npub                  <= '1';
                en_SelMode                <= bdi_valid;
                if (zcount < NumOfNWords) then
                    bdi_ready_s           <='1';
                    wen_npub              <=bdi_valid_s;
                    en_wcnt               <=bdi_valid_s;
                    -- rst_wcnt           <=not(bdi_valid_s and zdone);
                    wen_Tdata             <= bdi_valid_s;
                else
                    sel_npub_din          <= '1';
                    wen_npub              <= '1';
                    en_wcnt               <= '1';
                    rst_wcnt              <= not zdone;
                    wen_Tdata             <= '1';
                end if;
                if (bdi_eoi = '1') then
                    bdi_eoi_next          <= '1';
                end if;

            when S_WAIT_AD=>
                bdi_eoi_next              <= '0';

            when S_LD_AD=>
                bdi_ready_s               <='1';
                wen_ad                    <=bdi_valid_s;
                en_wcnt                   <=bdi_valid_s;
                rst_wcnt                  <=not(bdi_valid_s and zdone);
                wen_Tdata                 <= bdi_valid_s;
                end_of_block_s            <= zdone;
                if (bdi_eot = '1') then
                    bdi_eot_next          <= '1';
                end if;
                if (bdi_eoi = '1') and (pad_block = '0') then
                    bdi_eoi_next          <= '1';
                end if;

            when S_WAIT_PT=>
                bdi_eot_next              <= '0';
                bdi_eoi_next              <= '0';

            when S_LD_PT=>
                bdi_ready_s               <='1';
                wen_data                  <=bdi_valid_s;
                en_wcnt                   <=bdi_valid_s;
                rst_wcnt                  <=not(bdi_valid_s and zdone);
                if (dout_SelMode = '0') then
                    wen_Tdata             <= bdi_valid_s;
                end if;
                if(zdone='1') then
                    en_rcnt               <= bdi_valid_s;
                end if;
                if (bdi_eot = '1') then
                    bdi_eot_next          <= '1';
                end if;

            when S_PROC_DATA=>
                wen_cdata                 <='1';
                en_wcnt                   <='1';
                if (dout_SelMode = '1') then
                    wen_Tdata             <='1';
                    if (zpdone = '1')and(partial_r='1') then
                        pad_blk_proc_next <= '1';
                    end if;
                end if;
                rst_wcnt                  <=not(zdone);

            when S_OUT_CT=>
                bdo_valid_internal        <= '1';
                bdo_type                  <= HDR_CT;
                end_of_block_s            <= zdone_ct ;
                en_wcnt                   <= bdo_ready;
                rst_wcnt                  <= not(bdo_ready and zdone_ct);

            when S_PROC_TAG=>
                sel_tag                   <= '1';
                wen_cdata                 <= '1';
                en_wcnt                   <= '1';
                rst_wcnt                  <= not(zdone);
                end_of_block_s            <= zdone;

            when S_OUT_TAG=>
                bdo_valid_internal        <= '1';
                en_wcnt                   <= bdo_ready;
                end_of_block_s            <= zdone;
                rst_wcnt                  <= not(bdo_ready and zdone);
                bdo_type                  <= HDR_TAG;

            when S_VERIFY_TAG=>
                bdi_ready_s               <='1';
                en_wcnt                   <= bdi_valid;
                if (bdi_valid='1')then
                    if (bdi /= dout_tag)then
                        tag_error_next  <= '1';
                    end if;
                end if;
                if(zdone='1') then
                    if ((tag_error_r = '1') or (bdi /= dout_tag)) then
                        msg_auth_valid     <= '1';
                    else
                        msg_auth           <= '1';
                        msg_auth_valid     <= '1';
                    end if;
                end if;

            when S_DONE =>
                --done <='1';
            when others=>

        end case;
    end process;
    --=========================================================================
    WCNT: UpCountRst
      GENERIC MAP(addrWidth)
      PORT MAP     (
                        clk   => clk,
                        rst   => rst_wcnt,
                        en    => en_wcnt,
                        count => wrd_count
                    );
    WCNT2: UpCountRst
      GENERIC MAP(32)
      PORT MAP     (
                        clk   => clk,
                        rst   => rst_rcnt,
                        en    => en_rcnt,
                        count => rnd_count
                    );
    SelMode:   RegLd
                port map (
                        clk               => clk,
                        ena               => en_SelMode,
                        len               => len_SelMode,
                        load              => '0',
                        din               => decrypt_in,
                        dout              => dout_SelMode
                        );
    decrypt_out <= dout_SelMode;
    -- WCNT3: UpCountRst
      -- GENERIC MAP(addrWidth)
      -- PORT MAP     (
                        -- clk   => clk,
                        -- rst   => rst_tcnt,
                        -- en    => en_tcnt,
                        -- count => t_count
                    -- );
    addr_kram   <= wrd_count;
    addr_npub   <= wrd_count;
    addr_ad     <= wrd_count;
    addr_data   <= wrd_count;
    addr_cdata  <= wrd_count;
    addr_Tdata  <= wrd_count;
    zcount      <= to_integer(unsigned(wrd_count));
    --ercount       <= to_integer(unsigned(t_count));
    zdone       <= '1' when zcount = (NumOfWords-1) else '0';
    ndone       <= '1' when zcount = (NumOfNWords-1) else '0';
    zpdone      <= '1' when zcount = to_integer(unsigned(partial_blk_size)) else '0';
    zdone_ct    <= zpdone when partial_r='1' else zdone;
    bdo_valid   <= bdo_valid_internal;

end structure;
